----------------------------------------------------------------------------------
-- Author:         Armin Diehl

-- Create Date:    12/11/2006
-- Last Change:    12/22/2006 AD 
-- Module Name:    gide - Behavioral 
-- Project Name:   ecb ide interface
-- Target Devices: xc9572
-- Tool versions:  
-- Description:    ECB IDE Interface based on the idea of Tillmann Reh's GIDE
--                 The original design was implemented using PAL's and latches
--                 Needs the fitter option "Optimize Density" otherwise the fitter
--                 will not work (Unable to map all desired signals into function block)
-- Revision        0.03 - works (but has sometimes spikes on IDE_A0
--                 0.04 - Latched the addresses to the ide disk
--                 0.05 - Latched IDE_WR
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

-- for debug, Pin18 and Pin50 can be used

ENTITY gide IS
    Port ( ECBRes   : IN    STD_LOGIC;
           A        : IN    STD_LOGIC_VECTOR (7 downto 0);       -- ECB Address Bus
           JUMPER_A : IN    STD_LOGIC_VECTOR (3 downto 0);       -- upper 4 bit Jumper for IO Address
           IORQ     : IN    STD_LOGIC;
           RD       : IN    STD_LOGIC;
           WR       : IN    STD_LOGIC;
           D        : INOUT STD_LOGIC_VECTOR (7 downto 0);       -- ECB data bus
           CLK      : IN    STD_LOGIC;                           -- ECB clock
           IDE_D    : INOUT STD_LOGIC_VECTOR (15 downto 0);      -- 16 Bit data bus to/from ide
           IDE_CS0  : OUT   STD_LOGIC;                           -- CS0 to ide
           IDE_CS1  : OUT   STD_LOGIC;                           -- CS1 to ide
           IDE_RD   : OUT   STD_LOGIC;                           -- /IORD to ide
           IDE_WR   : OUT   STD_LOGIC;                           -- /IOWR to ide
           IDE_A    : OUT   STD_LOGIC_VECTOR (2 downto 0)        -- A0,A1,A2 to ide
           );
END gide;

architecture Behavioral OF gide IS
SIGNAL DOUT        : STD_LOGIC_VECTOR(7 downto 0);
SIGNAL DIDEOUT     : STD_LOGIC_VECTOR(15 downto 0);
SIGNAL CS_READ     : STD_LOGIC;
SIGNAL CS          : STD_LOGIC;
SIGNAL CS_16BIT    : STD_LOGIC;
SIGNAL WAS_16BIT   : STD_LOGIC;
SIGNAL READ_ACTIVE : STD_LOGIC;
SIGNAL READ_BUFFER : STD_LOGIC_VECTOR(7 downto 0);
SIGNAL CS_WRITE    : STD_LOGIC;
SIGNAL ACCESS_HIGH : STD_LOGIC;
SIGNAL ADDRESSOK   : STD_LOGIC;
SIGNAL SELECTDISK_WR: STD_LOGIC;
SIGNAL READSTATE : INTEGER RANGE 0 TO 2;

SIGNAL TOGGLERSTATE : INTEGER RANGE 0 TO 2;
SIGNAL IDE_RD_INT   : STD_LOGIC;

BEGIN
  -- address comparator, needed for IDECS0 and 1
  ADDRESSOK <= '1' WHEN A (7 downto 4) = JUMPER_A ELSE '0';
  
  --CS <= '1' WHEN (IORQ = '0') AND (ADDRESSOK = '1') AND ((RD = '0') OR (WR = '0')) ELSE '0';
  CS <= '1' WHEN (IORQ = '0') AND (ADDRESSOK = '1') ELSE '0';
  
  CS_16BIT <= '1' WHEN (CS = '1') AND (A(3 DOWNTO 0) = "1000") ELSE '0';
  
  -- generate CS_READ that will be 1 if IO read with matched address is active
  CS_READ <= '1' WHEN (CS = '1') AND (RD = '0') ELSE '0';
  
  -- generate CS_WRITE that will be 1 if IO write with matched address is active
  -- will activate process "writeData"
  CS_WRITE <= '1' WHEN (CS = '1') AND (WR = '0') ELSE '0';
  
  -- IDE_WR only on second access (16 Bit) or on any access (8 Bit)
  --IDE_WR <= '0' WHEN (SELECTDISK_WR = '1') AND (CS = '1') AND ((ACCESS_HIGH = '1') OR (CS_16BIT = '0')) ELSE '1';

  LatchIDE_WR:PROCESS(SELECTDISK_WR,WR)
  BEGIN
    IF WR = '1' THEN
      IDE_WR <= '1';
    ELSIF Rising_Edge(SELECTDISK_WR) THEN
      IDE_WR <= '0';
    END IF;
  END Process;

  -- Always set A0..A2 and CS0 and CS1 to the drive
  -- this gives spikes, IDE_A(0) will sometimes flip during writes
  --IDE_CS0 <= NOT A(3);
  --IDE_CS1 <= A(3);
  --IDE_A <= A(2 downto 0);
  
  LatchAddr:PROCESS(ECBRES,CLK)
  BEGIN
    IF (ECBRES = '0') THEN
      IDE_CS0 <= '0';
      IDE_CS1 <= '0';
      IDE_A <= "000";
    ELSIF Falling_Edge(CLK) THEN
      -- the Z80 address is valid short after a rising edge, latch it on falling edge
      IDE_CS0 <= NOT A(3);
      IDE_CS1 <= A(3);
      IDE_A <= A(2 downto 0);
    END IF;
  END PROCESS;
  
  -- flop flop for IDE_RD to avoid glitches
  LatchIdeRd:PROCESS(IDE_RD_INT,IORQ)
  BEGIN
    IF IORQ = '1' THEN
      IDE_RD <= '1';
    ELSIF Falling_Edge(IDE_RD_INT) THEN
      IDE_RD <= '0';
    END IF;
  END PROCESS;
  

  -- swaps ACCESS_HIGH for 16 bit accesses after the IO read or write has been completed
  -- do the toggle on rising edge after the IO has been completed
  HighLowToggler:PROCESS(CLK,ECBRes,CS)
  BEGIN
    IF ECBRes = '0' THEN
      ACCESS_HIGH <= '0';
      TOGGLERSTATE <= 0;
      WAS_16BIT <= '0';
    ELSIF Rising_Edge(CLK) THEN
        CASE TOGGLERSTATE IS
          WHEN 0 => IF CS = '1' THEN
                      TOGGLERSTATE <= 1;              -- IO now active
                      IF A(3 DOWNTO 0) = "1000" THEN
                        WAS_16BIT <= '1';
                      ELSE
                        WAS_16BIT <= '0';
                      END IF;
                    END IF;
          WHEN 1 => IF CS = '0' THEN                  -- IO finished ?
                      TOGGLERSTATE <= 2;              -- toggle the state on next clock
                    END IF;
          WHEN 2 => IF WAS_16BIT = '1' THEN           -- was it an access to 16 bit data ?
                      ACCESS_HIGH <= NOT ACCESS_HIGH; -- then toggle
                    ELSE
                      ACCESS_HIGH <= '0';             -- otherwise lower byte
                    END IF;
                    TOGGLERSTATE <= 0;
        END CASE;
    END IF;
  END PROCESS;
  
  
  writeData:PROCESS(CS_WRITE)          -- sets Data to the IDE drive depending
  BEGIN                                -- on 8 or 16 bit access
    IF Rising_Edge(CS_WRITE) THEN
      IF A(3 DOWNTO 0) = "1000" THEN   -- 16 Bit Access
        IF ACCESS_HIGH = '0' THEN      -- first byte
          DIDEOUT(7 downto 0) <= D;
        ELSE
          DIDEOUT(15 downto 8) <= D;   -- second byte
        END IF;
      ELSE
        DIDEOUT(15 downto 8) <= "00000000";  -- not really needed
        DIDEOUT(7 downto 0) <= D;
      END IF;
    END IF;
  END PROCESS;


    -- This sets the RW signal to the disk at a time when the
    -- Data to the IDE drive is valid
    SetWriteActive:PROCESS(CLK,ECBRes,CS_WRITE,ACCESS_HIGH)
    BEGIN
      IF ECBRes = '0' THEN
        SELECTDISK_WR <= '0';
      ELSIF Falling_Edge(CLK) THEN
        IF CS_WRITE = '1' THEN                          -- IOWR to 8 or 16 bit register
          IF A(3 DOWNTO 0) = "1000" THEN                -- access to data register (16 Bit)
            IF ACCESS_HIGH = '1' THEN                   -- if second byte is written
              SELECTDISK_WR <= '1';                     -- Enable /IOWR to ide
            ELSE
              SELECTDISK_WR <= '0';                     -- no disk access for first byte
            END IF;
          ELSE                                          -- access to normal register
            SELECTDISK_WR <= '1';                       -- Enable /IOWR to ide
          END IF;
        ELSE                                            -- falling edge of clk with no sctive write
          SELECTDISK_WR <= '0';                         -- Disable /IOWR to ide
        END IF;
      END IF;
    END PROCESS;
    
  
    SetReadActive:PROCESS(CLK,ECBRes)                   -- toggle CS_READ sync to clock
    BEGIN                                               -- to be able to read the ide data some time
      IF ECBRes = '0' THEN
        READ_Active <= '0';
        IDE_RD_INT <= '1';
        READSTATE <= 0;
      ELSIF Falling_Edge(CLK) THEN                      -- after CS0/CS1 are set
        IF CS_READ = '1' THEN
          IF READ_Active = '0' THEN
            READ_Active <= '1';
          END IF;
          IF READSTATE = 0 THEN                         -- first falling edge of clock in read
            IF (ACCESS_HIGH = '0') OR (CS_16BIT = '0') THEN
              IDE_RD_INT <= '0';                            -- enable IDE read
            ELSE
              IDE_RD_INT <= '1';
            END IF;
            READSTATE <= READSTATE + 1;
          ELSIF READSTATE = 1 THEN                      -- second falling edge of read
            IDE_RD_INT <= '1';                              -- disable read, we got the data on RisingEdge
            READSTATE <= READSTATE + 1;
          END IF;
        ELSE
          READ_Active <= '0';
          IDE_RD_INT <= '1';
          READSTATE <= 0;
        END IF;
      END IF;
    END PROCESS;
    
    -- on first access
    --  Output lower 8 bit to host and save upper 8 bit
    -- on second access output saved upper 8 bit
    READER:PROCESS(CLK,CS_16BIT,ACCESS_HIGH)
    BEGIN
      IF Rising_Edge(CLK) THEN
        IF READSTATE = 1 THEN                            -- indicates /READ was set on prev falling clock edge
          IF CS_16BIT = '1' THEN                         -- access to data register
            IF ACCESS_HIGH = '0' THEN                    -- read from IDE
              DOUT <= IDE_D(7 DOWNTO 0);                 -- lower 8 bit to data bus
              READ_BUFFER <= IDE_D(15 DOWNTO 8);         -- and save upper 8 bit
            ELSE
              DOUT <= READ_BUFFER;                       -- 2nd access, upper 8 bit to host
            END IF;
          ELSE                                           -- not data register
            DOUT <= IDE_D(7 DOWNTO 0);                   -- and output lower byte
          END IF;
        END IF;
      END IF;
    END PROCESS;
  
  -- set host data bus
  D <= DOUT WHEN (CS_READ = '1') ELSE "ZZZZZZZZ";
  
  -- set IDE data bus
  IDE_D <= DIDEOUT WHEN (CS_WRITE = '1') ELSE "ZZZZZZZZZZZZZZZZ";
  
END Behavioral;

